r"""Simple energy calculation using a double factorized hamiltonian."""

# imports
from pytket.extensions.qiskit import AerBackend
from pytket.partition import PauliPartitionStrat

from inquanto.express import get_system
from inquanto.ansatzes import (
    FermionSpaceAnsatzUCCSD,
    ComposedAnsatz,
    CircuitAnsatz,
    restricted_basis_rotation_to_circuit,
)
from inquanto.mappings import QubitMappingJordanWigner
from inquanto.protocols import PauliAveraging
from inquanto.computables.composite import ExpectationValueSumComputable

# Note: for a small system like this, we don't expect double factorization reduce the number of circuits
ham, space, state = get_system("h2_631g.h5")

# Truncate 2e integrals such that every element is within 1e-4 of exact val
df_ham = ham.double_factorize(tol1=1e-4, tol2=-1, method="cho")
jw = QubitMappingJordanWigner()

# Test a single ansatz configuration (random parameters)
ansatz = FermionSpaceAnsatzUCCSD(space, state, jw)
params = ansatz.state_symbols.construct_random(seed=1)
backend = AerBackend()

# Perform basis rotations, and map to qubit operators for each hamiltonian term
# (see https://inquanto.quantinuum.com/manual/spaces.html#double-factorization)
states, kernels = [], []
for operator, rotation in df_ham.items():
    rotator_circ = CircuitAnsatz(restricted_basis_rotation_to_circuit(rotation.T))
    rotated_ansatz = ComposedAnsatz(rotator_circ, ansatz)
    states.append(rotated_ansatz)
    kernels.append(operator.qubit_encode(jw))

# Build circuits for measuring total energy. Note that PauliAveraging supports a single ansatz, so we use the
# build_protocols_from() method to collect all circuits into a ProtocolList, with one protocol for each rotated ansatz.
energy_computable = ExpectationValueSumComputable(states, kernels)
protocols = PauliAveraging.build_protocols_from(
    params,
    energy_computable,
    backend=backend,
    shots_per_circuit=10000,
    pauli_partition_strategy=PauliPartitionStrat.CommutingSets,
)

# Run circuits and calculate energy
protocols.run(seed=1)
e = energy_computable.evaluate(protocols.get_evaluator())
print("Total energy: ", e)  # 0.9228419996965649
print("Num circs: ", protocols.n_circuit)
print("Circ depth: ", [c.depth() for c in protocols.get_circuits()])
